@@ -25,33 +25,16 @@ _VM_IMAGE_REGEX = re.compile(r"^[^\-]+-[^\-]+-[^\-]+$")
 _BACKEND_IMAGE_REGEX = re.compile(r"^[^\-]+$")
 
 
-def _storage_safe_encode(name: str) -> str:
+def storage_safe_encode(name: str) -> str:
     # Use URL quote so the names are still somewhat readable in the filesystem, but
     # we can unambiguously get the true name back for display purposes
     return urllib.parse.quote(name, safe="").replace("-", "%2D")
 
 
-def _storage_safe_decode(name: str) -> str:
+def storage_safe_decode(name: str) -> str:
     return urllib.parse.unquote(name)
 
 
-def _prepare_file_operation_bar(filesize: int) -> progressbar.ProgressBar:
-    return progressbar.ProgressBar(
-        maxval=filesize,
-        widgets=[
-            progressbar.Percentage(),
-            " ",
-            progressbar.Bar(),
-            " ",
-            progressbar.FileTransferSpeed(),
-            " | ",
-            progressbar.DataSize(),
-            " | ",
-            progressbar.ETA(),
-        ],
-    )
-
-
 class BaseImageProtocol:
     def __init__(self, regex: Pattern[str]) -> None:
         self.regex = regex
@@ -62,7 +45,9 @@ class BaseImageProtocol:
     def retrieve_image(
         self, store: "ImageStore", spec: "ImageSpec", destination: str
     ) -> None:
-        temp_destination = destination + ".part"
+        # Do partial downloads into the working directory in the backend
+        dest_name = os.path.basename(destination)
+        temp_destination = os.path.join(store.working, dest_name)
         fd = self.__lock_backend_destination(temp_destination)
 
         # We now hold the lock. Either another process started the retrieval
@@ -85,17 +70,6 @@ class BaseImageProtocol:
     ) -> None:
         raise RuntimeError("Protocol did not implement '_do_retrieve_image'")
 
-    def _copy_with_progress(
-        self, bar: progressbar.ProgressBar, source: IO[bytes], destination: IO[bytes]
-    ) -> None:
-        for idx in itertools.count():
-            block = source.read(_BLOCK_TRANSFER_SIZE)
-            if not block:
-                break
-            destination.write(block)
-            bar.update(idx * _BLOCK_TRANSFER_SIZE)
-        bar.finish()
-
     def __lock_backend_destination(self, dest: str) -> int:
         # By default, python 'open' call will truncate writable files. We can't allow that
         # as we don't yet hold the flock (and there is no way to open _and_ flock in one
@@ -183,7 +157,7 @@ class VagrantImageProtocol(BaseImageProtocol):
         box_file = tempfile.TemporaryFile()
 
         # Do the actual download
-        bar = _prepare_file_operation_bar(total_length)
+        bar = utils.prepare_file_operation_bar(total_length)
         for idx, block in enumerate(stream.iter_content(_BLOCK_TRANSFER_SIZE)):
             box_file.write(block)
             bar.update(idx * _BLOCK_TRANSFER_SIZE)
@@ -202,8 +176,7 @@ class VagrantImageProtocol(BaseImageProtocol):
             in_stream = tar.extractfile(image_info.name)
             assert in_stream is not None
 
-            bar = _prepare_file_operation_bar(image_info.size)
-            self._copy_with_progress(bar, in_stream, destination)
+            utils.copy_with_progress(in_stream, destination, image_info.size)
 
         logging.info("Image extraction completed.")
 
@@ -227,12 +200,31 @@ class FrontendImageProtocol(BaseImageProtocol):
             raise RuntimeError(f"Ambiguous backend image '{source}' for VM '{vm_name}'")
         frontend_image = candidates[0]
         with open(frontend_image.path, "rb") as existing_file:
-            bar = _prepare_file_operation_bar(frontend_image.actual_size)
-            self._copy_with_progress(bar, existing_file, destination)
+            utils.copy_with_progress(
+                existing_file, destination, frontend_image.actual_size
+            )
 
         logging.info("Image copy complete.")
 
 
+class FileImageProtocol(BaseImageProtocol):
+    def __init__(self) -> None:
+        super().__init__(re.compile(r"file", re.IGNORECASE))
+
+    def _do_retrieve_image(
+        self, store: "ImageStore", spec: "ImageSpec", destination: IO[bytes]
+    ) -> None:
+
+        print(f"Copying '{spec.source}' as new backend '{spec.name}'")
+
+        with open(spec.source, "rb") as existing_file:
+            size = existing_file.seek(0, os.SEEK_END)
+            existing_file.seek(0)
+            utils.copy_with_progress(existing_file, destination, size)
+
+        logging.info("File copy complete.")
+
+
 class HttpImageProtocol(BaseImageProtocol):
     def __init__(self) -> None:
         super().__init__(re.compile(r"http", re.IGNORECASE))
@@ -251,7 +243,7 @@ class HttpImageProtocol(BaseImageProtocol):
         if "content-length" in stream.headers:
             total_length = int(stream.headers["content-length"])
 
-        bar = _prepare_file_operation_bar(total_length)
+        bar = utils.prepare_file_operation_bar(total_length)
         for idx, block in enumerate(stream.iter_content(_BLOCK_TRANSFER_SIZE)):
             destination.write(block)
             bar.update(idx * _BLOCK_TRANSFER_SIZE)
@@ -265,6 +257,7 @@ _IMAGE_PROTOCOLS = [
     VagrantImageProtocol(),
     FrontendImageProtocol(),
     HttpImageProtocol(),
+    FileImageProtocol(),
 ]
 
 
@@ -302,9 +295,10 @@ class BaseImageInfo:
     image_info: Dict[str, Any]
 
     def __init__(self, store: "ImageStore", path: str) -> None:
-        stdout = subprocess.check_output(
+        stdout, _ = utils.run_check_retcode(
             [store.qemu_img_bin, "info", "-U", "--output=json", path]
         )
+        assert stdout is not None
         self.image_info = json.loads(stdout)
         self.store = store
         self.virtual_size = self.image_info["virtual-size"]
@@ -319,7 +313,7 @@ class BackendImageInfo(BaseImageInfo):
 
     def __init__(self, store: "ImageStore", path: str) -> None:
         super().__init__(store, path)
-        self.identifier = _storage_safe_decode(self.filename)
+        self.identifier = storage_safe_decode(self.filename)
 
 
 class FrontendImageInfo(BaseImageInfo):
@@ -330,12 +324,12 @@ class FrontendImageInfo(BaseImageInfo):
     def __init__(self, store: "ImageStore", path: str):
         super().__init__(store, path)
         vm_name, number, image = self.filename.split("-")
-        self.vm_name = _storage_safe_decode(vm_name)
+        self.vm_name = storage_safe_decode(vm_name)
         self.disk_number = int(number)
         backend_path = self.image_info["full-backing-filename"]
         try:
             self.backend = BackendImageInfo(store, backend_path)
-        except subprocess.CalledProcessError:
+        except utils.TransientProcessError:
             # If the path doesn't exist, it was either deleted before
             # we could run qemu-img, or it never existed at all
             if not os.path.exists(backend_path):
@@ -409,6 +403,7 @@ def format_image_table(
 class ImageStore:
     backend: str
     frontend: str
+    working: str
     qemu_img_bin: str
 
     def __init__(
@@ -417,6 +412,7 @@ class ImageStore:
 
         self.backend = os.path.abspath(backend_dir or self.__default_backend_dir())
         self.frontend = os.path.abspath(frontend_dir or self.__default_frontend_dir())
+        self.working = self.__working_dir()
         self.qemu_img_bin = self.__default_qemu_img_bin()
 
         if not os.path.exists(self.backend):
@@ -427,6 +423,14 @@ class ImageStore:
             logging.debug(f"Creating missing ImageStore frontend at '{self.frontend}'")
             os.makedirs(self.frontend, exist_ok=True)
 
+        if not os.path.exists(self.working):
+            os.makedirs(self.working, exist_ok=True)
+
+    def __working_dir(self) -> str:
+        # Note that the working directory must be in the same filesystem as
+        # the backend, to allow atomic movement of files.
+        return os.path.join(self.backend, ".working")
+
     def __default_backend_dir(self) -> str:
         env_specified = os.getenv("TRANSIENT_BACKEND")
         if env_specified is not None:
@@ -453,10 +457,13 @@ class ImageStore:
         else:
             raise RuntimeError(f"Invalid image file name: '{filename}'")
 
-    def retrieve_image(self, image_spec: str, vm_name: str) -> BackendImageInfo:
+    def backend_path(self, spec: ImageSpec) -> str:
+        safe_name = storage_safe_encode(spec.name)
+        return os.path.join(self.backend, safe_name)
+
+    def retrieve_image(self, image_spec: str) -> BackendImageInfo:
         spec = ImageSpec(image_spec)
-        safe_name = _storage_safe_encode(spec.name)
-        destination = os.path.join(self.backend, safe_name)
+        destination = self.backend_path(spec)
 
         if os.path.exists(destination):
             logging.info(f"Image '{spec.name}' already exists. Skipping retrieval")
@@ -472,9 +479,9 @@ class ImageStore:
     def create_vm_image(
         self, image_spec: str, vm_name: str, num: int
     ) -> FrontendImageInfo:
-        backing_image = self.retrieve_image(image_spec, vm_name)
-        safe_vmname = _storage_safe_encode(vm_name)
-        safe_image_identifier = _storage_safe_encode(backing_image.identifier)
+        backing_image = self.retrieve_image(image_spec)
+        safe_vmname = storage_safe_encode(vm_name)
+        safe_image_identifier = storage_safe_encode(backing_image.identifier)
         new_image_path = os.path.join(
             self.frontend, f"{safe_vmname}-{num}-{safe_image_identifier}"
         )
@@ -487,7 +494,7 @@ class ImageStore:
             f"Creating VM Image '{new_image_path}' from backing image '{backing_image.path}'"
         )
 
-        subprocess.check_output(
+        utils.run_check_retcode(
             [
                 self.qemu_img_bin,
                 "create",
@@ -507,12 +514,12 @@ class ImageStore:
     ) -> List[FrontendImageInfo]:
         images = []
         for candidate in os.listdir(self.frontend):
-            if not _VM_IMAGE_REGEX.match(candidate):
-                continue
             path = os.path.join(self.frontend, candidate)
+            if not os.path.isfile(path) or not _VM_IMAGE_REGEX.match(candidate):
+                continue
             try:
                 image_info = FrontendImageInfo(self, path)
-            except subprocess.CalledProcessError:
+            except utils.TransientProcessError:
                 # If the path doesn't exist anymore, then we raced with something
                 # that deleted the file, so just continue
                 if not os.path.exists(path):
@@ -536,12 +543,12 @@ class ImageStore:
     ) -> List[BackendImageInfo]:
         images = []
         for candidate in os.listdir(self.backend):
-            if not _BACKEND_IMAGE_REGEX.match(candidate):
-                continue
             path = os.path.join(self.backend, candidate)
+            if not os.path.isfile(path) or not _BACKEND_IMAGE_REGEX.match(candidate):
+                continue
             try:
                 image_info = BackendImageInfo(self, path)
-            except subprocess.CalledProcessError:
+            except utils.TransientProcessError:
                 if not os.path.exists(path):
                     continue
                 else:
