@@ -1,3 +1,5 @@
+from . import configuration
+from . import editor
 from . import qemu
 from . import image
 from . import utils
@@ -5,7 +7,7 @@ from . import ssh
 from . import sshfs
 from . import static
 
-import argparse
+import contextlib
 import enum
 import glob
 import logging
@@ -13,9 +15,10 @@ import os
 import pwd
 import signal
 import subprocess
+import tempfile
 import uuid
 
-from typing import cast, Optional, List, Dict, Any, Union, Tuple, TYPE_CHECKING
+from typing import cast, Iterator, Optional, List, Dict, Any, Union, Tuple, TYPE_CHECKING
 
 # _Environ is declared as generic in stubs but not at runtime. This makes it
 # non-subscriptable and will result in a runtime error. According to the
@@ -25,16 +28,6 @@ if TYPE_CHECKING:
 else:
     Environ = os._Environ
 
-# Maximum time to run virt-copy-in/out before failing
-_MAX_COPY_TIME = 500
-
-
-class TransientProcessError(Exception):
-    returncode: int
-
-    def __init__(self, returncode: int):
-        self.returncode = returncode
-
 
 @enum.unique
 class TransientVmState(enum.Enum):
@@ -45,13 +38,13 @@ class TransientVmState(enum.Enum):
 
 class TransientVm:
     store: image.ImageStore
-    config: argparse.Namespace
+    config: configuration.Config
     vm_images: List[image.FrontendImageInfo]
     ssh_config: Optional[ssh.SshConfig]
     qemu_runner: Optional[qemu.QemuRunner]
     qemu_should_die: bool
 
-    def __init__(self, config: argparse.Namespace, store: image.ImageStore) -> None:
+    def __init__(self, config: configuration.Config, store: image.ImageStore) -> None:
         self.store = store
         self.config = config
         self.vm_images = []
@@ -85,59 +78,6 @@ class TransientVm:
             or self.config.ssh_command is not None
         )
 
-    def __has_readable_kernel(self) -> bool:
-        kernels = glob.glob("/boot/vmlinuz-*")
-        if len(kernels) == 0:
-            logging.debug(f"No kernels found")
-            return False
-        for kernel in kernels:
-            if os.access(kernel, os.R_OK) is True:
-                logging.debug(f"Found readable kernel at {kernel}")
-                return True
-        logging.debug(f"No readable kernels found")
-        return False
-
-    def __extract_libguestfs_kernel(self) -> Tuple[str, str]:
-        """Extract the transient kernel and modules"""
-        home = utils.transient_data_home()
-        kernel_destination = os.path.join(home, "transient-kernel")
-        modules_destination = os.path.join(home, "transient-modules")
-        modules_dep = os.path.join(modules_destination, "modules.dep")
-        if os.path.exists(kernel_destination) and os.path.exists(modules_dep):
-            logging.debug(f"Transient kernel and modules already extracted. Skipping")
-            return kernel_destination, modules_destination
-
-        logging.info(f"Extracting transient kernel to {kernel_destination}")
-        utils.extract_static_file("transient-kernel", kernel_destination)
-
-        logging.info(f"Extracting transient modules to {modules_destination}")
-        # Our kernel doesn't actually use modules, but libguestfs expects _some_
-        # sort of modules directory, so just make an empty one
-        os.makedirs(modules_destination)
-        with open(modules_dep, "wb+") as f:
-            pass
-        return kernel_destination, modules_destination
-
-    def __prepare_libguestfs_environment(self) -> Environ:
-        """Configures an environment for running libguestfs commands"""
-        env = os.environ
-
-        # On platforms that don't have an available kernel or the kernel is
-        # not readable (e.g., docker and ubuntu), use ours.
-        if not self.__has_readable_kernel():
-            logging.info(f"No readable kernel detected. Extracting transient kernel")
-            kernel, modules = self.__extract_libguestfs_kernel()
-            env["SUPERMIN_MODULES"] = modules
-            env["SUPERMIN_KERNEL"] = kernel
-
-        # Avoid using libvirt explicitly
-        env["LIBGUESTFS_BACKEND"] = "direct"
-
-        # And do verbose logging so we can debug failures
-        env["LIBGUESTFS_TRACE"] = "1"
-        env["LIBGUESTFS_DEBUG"] = "1"
-        return env
-
     def __do_copy_command(self, cmd: List[str], environment: Environ) -> None:
         cmd_name = cmd[0]
         try:
@@ -148,13 +88,15 @@ class TransientVm:
                 stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE,
             )
-            _, raw_stderr = handle.communicate(timeout=_MAX_COPY_TIME)
+            _, raw_stderr = handle.communicate(timeout=self.config.copy_timeout)
             if handle.poll() == 0:
                 return
         except subprocess.TimeoutExpired:
             handle.terminate()
             _, raw_stderr = handle.communicate()
-            logging.error(f"{cmd_name} timed out after {_MAX_COPY_TIME} seconds")
+            logging.error(
+                f"{cmd_name} timed out after {self.config.copy_timeout} seconds"
+            )
         stderr = raw_stderr.decode("utf-8").strip()
         raise RuntimeError(f"{cmd_name} failed: {stderr}")
 
@@ -186,16 +128,14 @@ class TransientVm:
         if not vm_absolute_path.startswith("/"):
             raise RuntimeError(f"Absolute path for guest required: {vm_absolute_path}")
 
-        environment = self.__prepare_libguestfs_environment()
         for vm_image in self.vm_images:
             assert vm_image.backend is not None
             logging.info(
                 f"Copying from '{host_path}' to '{vm_image.backend.identifier}:{vm_absolute_path}'"
             )
-            self.__do_copy_command(
-                ["virt-copy-in", "-a", vm_image.path, host_path, vm_absolute_path],
-                environment,
-            )
+
+            with editor.ImageEditor(self.config, vm_image.path) as edit:
+                edit.copy_in(host_path, vm_absolute_path)
 
     def __needs_to_copy_out_files_after_running(self) -> bool:
         """Checks if at least one directory on the VM needs to be copied out
@@ -225,16 +165,14 @@ class TransientVm:
         if not vm_absolute_path.startswith("/"):
             raise RuntimeError(f"Absolute path for guest required: {vm_absolute_path}")
 
-        environment = self.__prepare_libguestfs_environment()
         for vm_image in self.vm_images:
             assert vm_image.backend is not None
             logging.info(
                 f"Copying from '{vm_image.backend.identifier}:{vm_absolute_path}' to '{host_path}'"
             )
-            self.__do_copy_command(
-                ["virt-copy-out", "-a", vm_image.path, vm_absolute_path, host_path,],
-                environment,
-            )
+
+            with editor.ImageEditor(self.config, vm_image.path) as edit:
+                edit.copy_out(vm_absolute_path, host_path)
 
     def __qemu_added_args(self) -> List[str]:
         new_args = ["-name", self.name]
@@ -283,9 +221,6 @@ class TransientVm:
     def __current_user(self) -> str:
         return pwd.getpwuid(os.getuid()).pw_name
 
-    def __qemu_guest_shutdown(self, event: qemu.QmpMessage) -> None:
-        logging.info(f"QEMU guest has shutdown. QMP event: {event}")
-
     def __qemu_sigchld_handler(self, sig: int, frame: Any) -> None:
         # We register this signal handler after the QEMU start, so these must not be None
         assert self.qemu_runner is not None
@@ -345,7 +280,7 @@ class TransientVm:
 
         if returncode != 0:
             logging.debug(f"VM exited with non-zero code: {returncode}")
-            raise TransientProcessError(returncode)
+            raise utils.TransientProcessError(returncode=returncode)
         return None
 
     def run(self) -> None:
@@ -363,7 +298,7 @@ class TransientVm:
         print("Finished preparation. Starting virtual machine")
 
         added_qemu_args = self.__qemu_added_args()
-        full_qemu_args = added_qemu_args + list(self.config.qemu_args)
+        full_qemu_args = added_qemu_args + self.config.qemu_args
 
         # If we are using the SSH console, we need to do _something_ with QEMU output.
         qemu_quiet, qemu_interactive = False, True
@@ -421,17 +356,9 @@ class TransientVm:
             # SIGCHLD exit.
             self.qemu_should_die = True
 
-            # If we get a guest SHUTDOWN signal, invoke a callback to log it
-            self.qemu_runner.qmp_client.register_callback(
-                "SHUTDOWN", self.__qemu_guest_shutdown
-            )
-
-            # Now actually request that the guest shutdown via ACPI
-            self.qemu_runner.qmp_client.send_sync({"execute": "system_powerdown"})
-
             try:
                 # Wait a bit for the guest to finish the shutdown and QEMU to exit
-                self.qemu_runner.wait(timeout=self.config.shutdown_timeout)
+                self.qemu_runner.shutdown(timeout=self.config.shutdown_timeout)
 
             except subprocess.TimeoutExpired:
                 # if the timeout == 0, then the user expects the guest to not actually
